use crate::bot::{Command, CommandContext};
use async_trait::async_trait;
use std::{net::IpAddr, str::FromStr};
use trust_dns_resolver::{
    config::{NameServerConfig, Protocol, ResolverConfig, ResolverOpts},
    name_server::{GenericConnection, GenericConnectionProvider, TokioRuntime},
    proto::rr::RecordType,
    AsyncResolver,
};

pub struct Dig {
    resolver: AsyncResolver<GenericConnection, GenericConnectionProvider<TokioRuntime>>,
}

struct Lookup {
    pub record_type: Option<RecordType>,
    pub domain: String,
    pub server: Option<IpAddr>,
}

impl Dig {
    pub fn new(alt_nserver: Option<String>) -> anyhow::Result<Self> {
        let resolver = if let Some(alt_nserver) = alt_nserver {
            let mut resolver_config = ResolverConfig::new();
            let nserver = format!("{alt_nserver}:53").parse()?;
            resolver_config.add_name_server(NameServerConfig::new(nserver, Protocol::Udp));
            AsyncResolver::tokio(resolver_config, ResolverOpts::default())?
        } else {
            AsyncResolver::tokio_from_system_conf()?
        };
        Ok(Self { resolver })
    }
}

#[async_trait]
impl Command for Dig {
    async fn execute(&self, msg: CommandContext) -> anyhow::Result<String> {
        if let Some(msg) = msg.content {
            let lookup = Lookup::parse(&msg)?;
            lookup.do_lookup(self).await
        } else {
            Ok("Usage: dig <record type> [hostname] @<dns server>".into())
        }
    }
}

impl Lookup {
    pub fn parse(input: &str) -> anyhow::Result<Self> {
        let parts: Vec<&str> = input.split_ascii_whitespace().collect();
        match parts.len() {
            1 => Ok(Self {
                record_type: None,
                domain: parts[0].to_string(),
                server: None,
            }),
            2 => {
                if parts[1].starts_with('@') {
                    let server = if let Some(server) = parts[1].strip_prefix('@') {
                        Some(server.parse()?)
                    } else {
                        None
                    };
                    Ok(Self {
                        record_type: None,
                        domain: parts[0].to_string(),
                        server,
                    })
                } else {
                    let record_type = Some(RecordType::from_str(&parts[0].to_ascii_uppercase())?);
                    Ok(Self {
                        record_type,
                        domain: parts[1].to_string(),
                        server: None,
                    })
                }
            }
            3 => {
                let record_type = Some(RecordType::from_str(&parts[0].to_ascii_uppercase())?);
                let server = if let Some(server) = parts[2].strip_prefix('@') {
                    Some(server.parse()?)
                } else {
                    None
                };
                Ok(Self {
                    record_type,
                    domain: parts[1].to_string(),
                    server,
                })
            }
            _ => anyhow::bail!("Usage: dig <record type> [hostname] @<dns server>"),
        }
    }

    pub fn parse_ptr(&self) -> anyhow::Result<String> {
        let addr = self.domain.parse::<IpAddr>();
        if let Some(RecordType::PTR) = self.record_type {
            let addr = addr?;
            if addr.is_ipv4() {
                let rev = self.domain.split('.').rev();
                let mut builder = String::new();
                for p in rev {
                    builder.push_str(&format!("{p}."));
                }
                builder.push_str("in-addr.arpa");
                Ok(builder)
            } else {
                let mut builder = String::new();
                if let IpAddr::V6(addr) = addr {
                    let mut octets = addr.octets();
                    octets.reverse();
                    for b in octets {
                        let str = format!("{b:02X?}");
                        let chars = str.chars().rev();
                        for c in chars {
                            builder.push(c);
                            builder.push('.');
                        }
                    }

                    builder.push_str("ip6.arpa");
                }
                Ok(builder)
            }
        } else {
            Ok(self.domain.clone())
        }
    }

    pub async fn do_lookup(&self, dig: &Dig) -> anyhow::Result<String> {
        let mut builder = String::new();
        if let Some(record_type) = self.record_type {
            let domain = self.parse_ptr()?;
            if let Some(server) = &self.server {
                let mut resolver_config = ResolverConfig::new();
                let nserver = if server.is_ipv4() {
                    format!("{}:53", server.to_string()).parse()?
                } else {
                    format!("[{}]:53", server.to_string()).parse()?
                };
                resolver_config.add_name_server(NameServerConfig::new(nserver, Protocol::Udp));
                let tmp_resover = AsyncResolver::tokio(resolver_config, ResolverOpts::default())?;
                let records = tmp_resover.lookup(domain.clone(), record_type).await?;
                for r in records {
                    builder.push_str(&r.to_string());
                }
            } else {
                let records = dig.resolver.lookup(domain.clone(), record_type).await?;
                for r in records {
                    builder.push_str(&r.to_string());
                }
            }
        } else if self.server.is_some() {
            let mut resolver_config = ResolverConfig::new();
            let nserver = format!("{}:53", self.server.as_ref().unwrap()).parse()?;
            resolver_config.add_name_server(NameServerConfig::new(nserver, Protocol::Udp));
            let tmp_resover = AsyncResolver::tokio(resolver_config, ResolverOpts::default())?;
            let records = tmp_resover.lookup_ip(&self.domain).await?;
            for r in records {
                builder.push_str(&r.to_string());
            }
        } else {
            let records = dig.resolver.lookup_ip(&self.domain).await?;
            for r in records {
                builder.push_str(&r.to_string());
            }
        }

        Ok(builder)
    }
}
