use crate::{
    bot::{Command, CommandContext},
    config::TraceConfig,
};
use async_trait::async_trait;
use std::sync::Arc;
use tokio::task::JoinSet;

pub struct Trace {
    cfg: Arc<TraceConfig>,
}

impl Trace {
    pub fn new(cfg: Arc<TraceConfig>) -> Self {
        Self { cfg }
    }
}

#[async_trait]
impl Command for Trace {
    async fn execute(&self, ctx: CommandContext) -> anyhow::Result<String> {
        if let Some(msg) = ctx.content {
            if msg.starts_with('-') || msg.contains(" -") {
                return Ok("Please don't try to pass weird commands to MTR :)".into());
            }

            let mut join_set = JoinSet::new();
            for server in &self.cfg.server {
                let url = format!(
                    "http://{}.{}:{}/traceroute?q={}",
                    server, self.cfg.base_url, self.cfg.port, msg
                );
                join_set.spawn(reqwest::get(url));
            }

            let mut send = String::new();
            while let Some(trace) = join_set.join_next().await {
                let trace = trace??;
                let server = trace.url().host().unwrap().to_string();
                let res = trace.text().await?;
                send.push_str(&parser(&res, &server));
                send.push_str("\r\n");
            }

            if send.is_empty() {
                return Ok("Trace failed".into());
            }

            return Ok(send);
        }

        Ok("No address specified".into())
    }
}

pub struct Trace6 {
    cfg: Arc<TraceConfig>,
}

impl Trace6 {
    pub fn new(cfg: Arc<TraceConfig>) -> Self {
        Self { cfg }
    }
}

#[async_trait]
impl Command for Trace6 {
    async fn execute(&self, ctx: CommandContext) -> anyhow::Result<String> {
        if let Some(msg) = ctx.content {
            if msg.starts_with('-') || msg.contains(" -") {
                return Ok("Please don't try to pass weird commands to MTR :)".into());
            }

            let mut join_set = JoinSet::new();
            for server in &self.cfg.server {
                let url = format!(
                    "http://{}.{}:{}/traceroute?q={}",
                    server, self.cfg.base_url, self.cfg.port, msg
                );
                join_set.spawn(reqwest::get(url));
            }

            let mut send = String::new();
            while let Some(trace) = join_set.join_next().await {
                let trace = trace??;
                let server = trace.url().host().unwrap().to_string();
                let res = trace.text().await?;
                send.push_str(&parser(&res, &server));
                send.push_str("\r\n");
            }

            if send.is_empty() {
                return Ok("Trace failed".into());
            }

            return Ok(send);
        }

        Ok("No address specified".into())
    }
}

fn parser(msg: &str, server: &str) -> String {
    let mut builder = String::new();
    let mut max_len = 400;
    if builder.starts_with("mtr") {
        return builder;
    }
    let mut lines = msg.lines();
    let _ = lines.nth(1);

    builder.push_str(&format!("{server} "));
    if lines.clone().next().is_none() {
        builder.push_str("| Route unreachable");
    }

    for l in lines {
        if l.starts_with("mtr") {
            return l.into();
        }
        let mut split = l.split_whitespace();

        let (Some(hop), Some(host), Some(ip), Some(avg)) = (split.next(), split.next(), split.next(), split.nth(3)) else {
            continue;
        };

        let nextword = if host == "???" {
            "| * ".into()
        } else {
            let hop = hop.split('|').next().unwrap_or("");
            if format!("({host})") == ip {
                format!("| {hop} {host} {avg}ms ")
            } else {
                format!("| {hop} {host} {ip} {avg}ms ")
            }
        };

        if builder.len() + nextword.len() > max_len {
            builder.push_str("\r\n");
            max_len *= 2;
        }

        builder.push_str(&nextword);
    }

    builder
}
